#include <AccelStepper.h>
#include <Arduino.h>
#include "wiring_private.h"

#ifndef _VARIANT_MINITRONICS_20_
  #error Please select board minitronics v2.0
#endif

// ############# PLEASE KEEP IN MIND: HALF STEPPING!!!! #############
// Full rotation normal stepping = 200 steps
// In half stepping, full rotation is 400 steps
// 360 degree rotation = 400 steps

#define DEBUG_MODE                0 // 1=on, 0=off

// Define target detection angle range
#define X_ANGLE_LOWER_LIMIT       -90 // °
#define X_ANGLE_UPPER_LIMIT       90 // °
#define DETECTION_THRESHOLD       3500 // mm
#define X_DETECTION_HYSTERESIS    2    // °degree, 5 degree is equal to 1.5%
#define Y_DETECTION_HYSTERESIS    1    // °degree, 3 degree is equal to around 1%
#define NO_TARGET_TIMEOUT         10 // timing depends on lidar scan frequency
#define X_AXIS_RANGE              180 // °degree
#define GEARBOX_RATIO             9 // 1:9

#define Y_ANGLE_LOWER_LIMIT       0 // °
#define Y_ANGLE_UPPER_LIMIT       90// °

// Stepper defines
#define STEPS_FULL_ROTATION       200 // 1.8 degree/step
#define Y_MOVING_DISTANCE         65 // mm that is equal to 90 degree movement

#define Y_MICROSTEPS              2 // 2 = half step
#define Y_STEPPER_SPEED           600
#define Y_STEPPER_ACCELERATION    2000

#define X_MICROSTEPS              1 // 1 = full step
#define X_STEPPER_SPEED           450
#define X_STEPPER_ACCELERATION    1000


#define LEADSCREW_VALUE           8 //mm per rotation

// Pin map
#define VALVE_A 25
#define VALVE_B 26
#define VALVE_C 27
#define VALVE_D 33

#define STEPPERX_DIR_PIN 19
#define STEPPERX_STEP_PIN 20
#define STEPPERX_ENABLE_PIN 21

#define STEPPERY_DIR_PIN 28
#define STEPPERY_STEP_PIN 29
#define STEPPER_ENABLE_PIN 0

#define ENDSWITCH_X 9
#define ENDSWITCH_Y 4
#define LED_PIN 13
#define PWM_PIN 24
#define RPLIDAR_MOTOR 7 // The PWM pin for control the speed of RPLIDAR's motor.
#define DEBUG_LED_A 6
#define DEBUG_LED_B 10
#define DEBUG_LED_C 11
#define POT_PIN_ANALOG 30
#define PIN_SERIAL2_RX       (45ul)
#define PIN_SERIAL2_TX       (44ul) 
#define PAD_SERIAL2_TX       (UART_TX_PAD_2)
#define PAD_SERIAL2_RX       (SERCOM_RX_PAD_3)

// Define some steppers and the pins the will use
AccelStepper StepperX(AccelStepper::DRIVER, STEPPERX_STEP_PIN, STEPPERX_DIR_PIN);
AccelStepper StepperY(AccelStepper::DRIVER, STEPPERY_STEP_PIN, STEPPERY_DIR_PIN);

Uart mySerial (&sercom5, PIN_SERIAL2_RX, PIN_SERIAL2_TX, PAD_SERIAL2_RX, PAD_SERIAL2_TX); // Create the new UART instance for the Lidar sensor

volatile unsigned char scan_state_flag = 0;
volatile unsigned char rxBuffer[5];
volatile unsigned char process_flag = 0;
volatile unsigned int sample_cntr = 0;
volatile unsigned char lidar_error_flag = 1; // default 1 to establish connection for the first time

volatile float current_angle = 0;
volatile float current_distance = 0;
volatile float alpha = 0.5; // filter constant
volatile float x_target_angle = 0; // filtered values
volatile float y_target_angle = 0; // calculated according target_distance
volatile float target_distance = 0; // filtered values
volatile float target_distance_detection = 100000; // raw values
volatile float target_angle_detection = 0; // raw values

volatile unsigned int fire_cntr = 0;

volatile float y_axis_coefficient = -((float)Y_ANGLE_UPPER_LIMIT*0.5)/(float)DETECTION_THRESHOLD;

//******************************************************************************************//
//                                                                                          //
//                                      RPLidar functions                                   //
//                                                                                          //
//******************************************************************************************//
unsigned int lidar_request_info() 
{
  const unsigned int rx_lenght = 43;
  unsigned char tx_buffer[2] = {0xA5, 0x5A};
  unsigned char rx_buffer[rx_lenght];
  unsigned int rx_cnt = 0;

  // flush rx buffer
  while(mySerial.available()){mySerial.read();}
  
  mySerial.write(tx_buffer, sizeof(tx_buffer));
  rx_cnt = mySerial.readBytes(rx_buffer, rx_lenght);
 
  if( (rx_cnt == rx_lenght) && (rx_buffer[0] == 0xA5) && (rx_buffer[1] == 0x5A) )
    return 1;
  else
    return 0;     
}

unsigned int lidar_start_scan() 
{
  const unsigned int rx_lenght = 7;
  unsigned char tx_buffer[2] = {0xA5, 0x20};
  unsigned int rx_cnt = 0;
  unsigned char rx_buffer[rx_lenght];
  
  mySerial.write(tx_buffer, sizeof(tx_buffer)); 

  rx_cnt = mySerial.readBytes(rx_buffer, rx_lenght);
  
  if( (rx_cnt == rx_lenght) && (rx_buffer[0] == 0xA5) && (rx_buffer[1] == 0x5A) && (rx_buffer[2] == 0x05) )
  {
    // start motor rotating at max allowed speed
    analogWrite(RPLIDAR_MOTOR, 255);         
    return 1;
  }
  else
    return 0;    
  
}

void lidar_stop_scan() 
{
  unsigned char tx_buffer[2] = {0xA5, 0x25};
  mySerial.write(tx_buffer, sizeof(tx_buffer));
  // stop motor
  analogWrite(RPLIDAR_MOTOR, 0);
}

//******************************************************************************************//
//                                                                                          //
//                                      UART ISR                                            //
//                                                                                          //
//******************************************************************************************//
void SERCOM5_Handler()
{
  if(scan_state_flag) // only start processing data if Lidar has been detected and started
  {    
    while(sercom5.availableDataUART()) 
    {
      for(int i = 0; i < 4; i++)
        rxBuffer[i] = rxBuffer[i+1];
      rxBuffer[4] = sercom5.readDataUART();
          
      if(sample_cntr < 4)
        sample_cntr++;       
      else
      {
        // 5 bytes received
        if(rxBuffer[1] & (1<<0)) // Checkbit check!
        {  
          if( (rxBuffer[0] & (1<<0)) && !(rxBuffer[0] & (1<<1)) ) // new start of 360 degree scan!
          {
            digitalWrite(DEBUG_LED_A, !digitalRead(DEBUG_LED_A)); // frequency counter
            // reset lowest angle
            target_distance = ( (target_distance_detection*alpha)+((1-alpha)*target_distance) );
            if(target_distance_detection == DETECTION_THRESHOLD+10)
              x_target_angle = 0;
            else
              x_target_angle = ( (target_angle_detection*alpha)+((1-alpha)*x_target_angle) );
            target_distance_detection = DETECTION_THRESHOLD+10; // always more than minimal threshold, reset value
            process_flag = 1;            
          }
               
          // Process data
          current_angle = ((rxBuffer[1]>>1)+(rxBuffer[2]*128))/64.0f;

          // Out of range
          if( (current_angle >= 90.0) && (current_angle <= 270.0) )
          {
            // Ignore this area
          }
          else // visible range!
          {
            // Convert to +90 and -90
            if( (current_angle >= 270.0) && (current_angle <= 360.0) )
              current_angle = current_angle-360.0;
              
            current_distance = (rxBuffer[3]+(256*rxBuffer[4]))/4.0f;        
            if( (current_distance < target_distance_detection) && (current_distance != 0) )
            {            
              target_angle_detection = -current_angle;
              target_distance_detection = current_distance;
            }
          }
          
        }
        else
        {
          // Reset VARS
          sample_cntr = 0;
          scan_state_flag = 0;
          lidar_error_flag = 1;
          SerialUSB.print("Error checking check bit!\n");
        }
        sample_cntr = 0;        
      }
    }
  }  
  mySerial.IrqHandler();

}

//******************************************************************************************//
//                                                                                          //
//                                      Disable gun                                         //
//                                                                                          //
//******************************************************************************************//
void disable_gun(void)
{ 
  if(DEBUG_MODE) SerialUSB.println("Disable gun!");
  digitalWrite(VALVE_A, HIGH);
  digitalWrite(VALVE_B, HIGH);
  digitalWrite(VALVE_C, HIGH);
  digitalWrite(VALVE_D, HIGH);
  //fire_cntr = 0; no delay, direct fire! Remember prev setting
}

//******************************************************************************************//
//                                                                                          //
//                                      Fire!                                               //
//                                                                                          //
//******************************************************************************************//
void combat_mode(void)
{ 
  // Fire in the hole!
  if(DEBUG_MODE) SerialUSB.println("Fire!");
  if(fire_cntr < 8)
    fire_cntr++;
  else
    fire_cntr = 0;

  if(DEBUG_MODE) SerialUSB.println(fire_cntr);
   
  switch(fire_cntr)
  {
    case 2:
      digitalWrite(VALVE_A, LOW);
      digitalWrite(VALVE_B, HIGH);
      digitalWrite(VALVE_C, HIGH);
      digitalWrite(VALVE_D, HIGH);      
      break;
    case 4:
      digitalWrite(VALVE_A, HIGH);
      digitalWrite(VALVE_B, LOW);
      digitalWrite(VALVE_C, HIGH);
      digitalWrite(VALVE_D, HIGH);      
      break;
    case 6:
      digitalWrite(VALVE_A, HIGH);
      digitalWrite(VALVE_B, HIGH);
      digitalWrite(VALVE_C, LOW);
      digitalWrite(VALVE_D, HIGH);      
      break;
    case 8:
      digitalWrite(VALVE_A, HIGH);
      digitalWrite(VALVE_B, HIGH);
      digitalWrite(VALVE_C, HIGH);
      digitalWrite(VALVE_D, LOW);
      break;                              
  }
}

//******************************************************************************************//
//                                                                                          //
//                                      init axis stepper                                   //
//                                                                                          //
//******************************************************************************************//
void init_axis(void)
{ 
 
  StepperX.moveTo(-x_angle_to_steps(X_AXIS_RANGE));  
  StepperY.moveTo(y_angle_to_steps(Y_ANGLE_UPPER_LIMIT)); // 90° = forward direction
  
  digitalWrite(STEPPER_ENABLE_PIN, LOW); // low = enable  
  digitalWrite(STEPPERX_ENABLE_PIN, LOW); // low = enable  
  
  while( (digitalRead(ENDSWITCH_X) && StepperX.isRunning()) || (digitalRead(ENDSWITCH_Y) && StepperY.isRunning()) ) // stop at end switch or steps limit
  {
    if(digitalRead(ENDSWITCH_Y) && StepperY.isRunning())
      StepperY.run();
    
    if(digitalRead(ENDSWITCH_X) && StepperX.isRunning())
      StepperX.run();  
  }     
  
  StepperX.setCurrentPosition(x_angle_to_steps(-90)); // this is the new 0-position, Todo: check if this is OK!?
  StepperY.setCurrentPosition((Y_MOVING_DISTANCE/LEADSCREW_VALUE)*STEPS_FULL_ROTATION*Y_MICROSTEPS); // this is the new 0-position = 3250
  
  digitalWrite(STEPPER_ENABLE_PIN, HIGH); // low = enable, high = disable
  digitalWrite(STEPPERX_ENABLE_PIN, HIGH); // low = enable, high = disable

  if(DEBUG_MODE)
    SerialUSB.println("Init completed!");
}

//******************************************************************************************//
//                                                                                          //
//                                      Convert Y axis to steps                             //
//                                                                                          //
//******************************************************************************************//
float y_angle_to_steps(float angle_value)
{ 
  // angle range: 0° - 45°
  // Y-moving 0mm - 65mm = 0° - 45°
  // LEADSCREW_VALUE = 8mm / rotation  
  // 65/8 = number of rotations = 8,125
  // 8.125 * STEPS_FULL_ROTATION = STEPS = 1625

  float mm_setpoint = ((float)Y_MOVING_DISTANCE/90)*angle_value;
  float result = (mm_setpoint/(float)LEADSCREW_VALUE)*(float)STEPS_FULL_ROTATION*(float)Y_MICROSTEPS;

  return result; 
  
}

//******************************************************************************************//
//                                                                                          //
//                                      Convert Y steps to angle                            //
//                                                                                          //
//******************************************************************************************//
float y_steps_to_angle(int steps)
{  
  float travel_mm = steps/(((float)STEPS_FULL_ROTATION*(float)Y_MICROSTEPS)/(float)LEADSCREW_VALUE); 
  float result = travel_mm/((float)Y_MOVING_DISTANCE/90);
    
  return result;  
}

//******************************************************************************************//
//                                                                                          //
//                                      Convert X angle to steps                            //
//                                                                                          //
//******************************************************************************************//
float x_angle_to_steps(float angle_value)
{ 
  float steps_per_degree = (((float)STEPS_FULL_ROTATION*(float)X_MICROSTEPS)/360); 
  float result = ((float)angle_value*steps_per_degree*(float)GEARBOX_RATIO); // 6 is the gearbox ratio;
  return result;  
}

//******************************************************************************************//
//                                                                                          //
//                                      Convert X steps to angle                            //
//                                                                                          //
//******************************************************************************************//
float x_steps_to_angle(int steps)
{ 
  float steps_per_degree = (((float)STEPS_FULL_ROTATION*(float)X_MICROSTEPS)/360); 
  float result = (steps/(float)GEARBOX_RATIO)/steps_per_degree; 
  return result;  
}

//******************************************************************************************//
//                                                                                          //
//                                      SETUP                                               //
//                                                                                          //
//******************************************************************************************//
void setup() 
{     
  // End switch init
  pinMode(ENDSWITCH_X, INPUT_PULLUP);
  pinMode(ENDSWITCH_Y, INPUT_PULLUP);
  
  // POT pin
  pinMode(POT_PIN_ANALOG, INPUT);

  //LEDS
  pinMode(DEBUG_LED_A, OUTPUT);
  digitalWrite(DEBUG_LED_A, LOW);
  pinMode(DEBUG_LED_B, OUTPUT);
  digitalWrite(DEBUG_LED_B, LOW);
  pinMode(DEBUG_LED_C, OUTPUT);
  digitalWrite(DEBUG_LED_C, LOW);  

  // Valves
  pinMode(VALVE_A, OUTPUT);
  pinMode(VALVE_B, OUTPUT);
  pinMode(VALVE_C, OUTPUT);
  pinMode(VALVE_D, OUTPUT);
  disable_gun();
  
  // set pin modes
  pinMode(RPLIDAR_MOTOR, OUTPUT);

  // PWM init 
  pinMode(PWM_PIN, OUTPUT);
  analogWrite(PWM_PIN, 128);
  //digitalWrite(PWM_PIN, LOW);
  
  SerialUSB.begin(115200); // debug
  mySerial.begin(115200);  // lidar

  StepperY.setMaxSpeed(Y_STEPPER_SPEED);
  StepperY.setAcceleration(Y_STEPPER_ACCELERATION);
     
  StepperX.setMaxSpeed(X_STEPPER_SPEED);
  StepperX.setAcceleration(X_STEPPER_ACCELERATION);
  pinMode(STEPPER_ENABLE_PIN, OUTPUT);
  pinMode(STEPPERX_ENABLE_PIN, OUTPUT);
  digitalWrite(STEPPER_ENABLE_PIN, HIGH); // low = enable, high = disable
  digitalWrite(STEPPERX_ENABLE_PIN, HIGH); // low = enable, high = disable
  
}

//******************************************************************************************//
//                                                                                          //
//                                        LOOP                                              //
//                                                                                          //
//******************************************************************************************//
void loop() 
{

  unsigned int no_target_cntr = 0;
  
  delay(1000);
  init_axis();
  delay(1000);
 
  while(1)
  {
    if(lidar_error_flag) // Handle lidar RX data error
    {
      lidar_stop_scan();
      if(lidar_request_info() == 0)
      {
        SerialUSB.print("Lidar not detected!\n");
        digitalWrite(STEPPER_ENABLE_PIN, LOW); // low = enable, high = disable
        digitalWrite(STEPPERX_ENABLE_PIN, LOW); // low = enable, high = disable
        delay(2000);
      } 
      else
      {   
        SerialUSB.print("Lidar detected!\n");
        digitalWrite(STEPPER_ENABLE_PIN, HIGH); // low = enable, high = disable
        digitalWrite(STEPPERX_ENABLE_PIN, HIGH); // low = enable, high = disable
        scan_state_flag = lidar_start_scan();
        lidar_error_flag = 0;
      }      
    }
    else
    {
      if(process_flag)
      {        
        if(DEBUG_MODE)
        {
          char buf[10];
          sprintf(buf, "Target distance: %.1f mm at angle: %.1f", target_distance, x_target_angle);
          SerialUSB.println(buf);
        }
        
        digitalWrite(STEPPER_ENABLE_PIN, LOW); // low = enable        
        digitalWrite(STEPPERX_ENABLE_PIN, LOW); // low = enable        

        if(target_distance < DETECTION_THRESHOLD) // Respond!
        {
          y_target_angle = ((target_distance*y_axis_coefficient)+65); // 65 is manually tuned value
          unsigned char x_axis_adjust = (x_target_angle>(x_steps_to_angle(StepperX.targetPosition())+(float)X_DETECTION_HYSTERESIS)) || (x_target_angle<(x_steps_to_angle(StepperX.targetPosition())-(float)X_DETECTION_HYSTERESIS));
          unsigned char y_axis_adjust = (y_target_angle>(y_steps_to_angle(StepperY.targetPosition())+(float)Y_DETECTION_HYSTERESIS)) || (y_target_angle<(y_steps_to_angle(StepperY.targetPosition())-(float)Y_DETECTION_HYSTERESIS));

          // First handle x axis (angle)!
          if(x_axis_adjust)
            StepperX.moveTo(x_angle_to_steps(x_target_angle));

          if(y_axis_adjust)
            StepperY.moveTo(y_angle_to_steps(y_target_angle));

          // determine if gun is into position and able to fire!
          unsigned char x_axis_in_range = (x_target_angle>(x_steps_to_angle(StepperX.currentPosition())-(float)X_DETECTION_HYSTERESIS)) && (x_target_angle<(x_steps_to_angle(StepperX.currentPosition())+(float)X_DETECTION_HYSTERESIS));
          unsigned char y_axis_in_range = (y_target_angle>(y_steps_to_angle(StepperY.currentPosition())-(float)Y_DETECTION_HYSTERESIS)) && (y_target_angle<(y_steps_to_angle(StepperY.currentPosition())+(float)Y_DETECTION_HYSTERESIS));
          
          // If lock on target, then fire shots!
          if(x_axis_in_range && y_axis_in_range)
            combat_mode();
          else
            disable_gun();       
          
          no_target_cntr = 0;
        }
        else
        {
          // Check if target is out of range for a while before shutting down system, then:
          // Goto default combat position, then disable steppers to save energy    
          if(no_target_cntr < NO_TARGET_TIMEOUT)
            no_target_cntr++;
          else
          {
            if(DEBUG_MODE) SerialUSB.println("No target detected for a while, goto idle position...");

            if(x_steps_to_angle(StepperX.currentPosition()) != 0)
              StepperX.moveTo(x_angle_to_steps(0));
              StepperY.moveTo(y_angle_to_steps(((float)Y_ANGLE_UPPER_LIMIT*0.5)));
            disable_gun();
          }        

        }       
        process_flag = 0;
      }      

      // This call should be made frequently
      StepperX.run();
      StepperY.run();

      
    }

  

    //delay(1000);
  }
  
 
}
